/*
  SNESAPU 
  copyright (C)2003-04 Alpha-II Productions
  http://www.alpha-ii.com/

  Ported from x86 assembler to C by Brad Jorsch

  Person who rewrites this.
  Copyright (C) 2004 sanmaiwashi
  http://toriniku.sourceforge.net/
*/

#include "snes9x.h"
#include "memmap.h"
#include "apu.h"
#include "soundux.h"
#include "snesapu.h"
#include <math.h>

/* For note-triggered SPC dump support */
#include "snapshot.h"

#ifdef __UOSNES__
extern const char *S9xGetSPCFilename ();
#else
extern "C" {const char *S9xGetFilenameInc (const char *);}
#endif

// Use pregenerated gaussian curve from SNES
//#define	NO_ESA
#define	SA_SNESINT

typedef float f32;
typedef double f64;
typedef long double f80;

#ifdef M_PIl
#undef M_PIl
#endif
#define M_PIl		3.1415926535897932384626433832795029L

// Note: if x>>n doesn't do an arithmetic shift when x is signed, you might
// want to consider overloading the >> operator, at least for x of types int8,
// int16, and int32...

/***********************************************************************
  Defines
***********************************************************************/

// Envelope mode masks
#define E_TYPE   0x01	  // Type of adj: Constant(1/64 or 1/256) / Exp.(255/256)
#define E_DIR    0x02     // Direction: Decrease / Increase
#define E_DEST   0x04     // Destination: Default(0 or 1) / Other(x/8 or .75)
#define E_ADSR   0x08     // Envelope mode: Gain/ADSR
#define E_IDLE   0x80     // Envelope speed is set to 0

#define E_DEC    0x00     // Linear decrease
#define E_EXP    0x01     // Exponential decrease
#define E_INC    0x02     // Linear increase
#define E_BENT   0x06     // Bent line increase
#define E_REL    0x08     // Release mode
#define E_SUST   0x09     // Sustain mode
#define E_ATT    0x0A     // Attack mode
#define E_DECAY  0x0D     // Decay mode
#define E_DIRECT 0x87     // Direct gain

// Envelope adjustment rates ------------------
#define A_GAIN   (1 << E_SHIFT)					// Amount to adjust envelope values
#define A_LIN    ((128*A_GAIN)/64)				// Linear rate to increase/decrease envelope
#define A_KOF    ((128*A_GAIN)/256)				// Rate to decrease envelope during release
#define A_BENT   ((128*A_GAIN)/256)				// Rate to increase envelope after bend
#define A_NOATT  (64*A_GAIN)					// Rate to increase if attack rate is set to 0ms
#define A_EXP    0								// Rate to decrease envelope exponentially (Not used)

// Envelope destination values ----------------
#define D_MAX    ((128*A_GAIN)-1)				// Maximum envelope value
#define D_ATTACK ((128*A_GAIN*63)/64)			// Destination of attack rate
#define D_BENT   ((128*A_GAIN*3)/4)				// First destination of bent line
#define D_EXP    ((128*A_GAIN)/8)				// Minimum decay destination value
#define D_SUST   ((128*A_GAIN)/10)              // Destination volume for exponential decrease
#define D_MIN    0								// Minimum envelope value

#undef ABS
#define ABS(a) ((a) < 0 ? -(a) : (a))

#define FIXED_POINT 0x10000UL
#define FIXED_POINT_REMAINDER 0xffffUL
#define FIXED_POINT_SHIFT 16

/***********************************************************************
  Variables
***********************************************************************/

static int32 brrTab[256];
static int16 cubicTab[1024];
static int16 gaussTab[1024];
#ifdef SA_SNESINT
#include "gausspre.h"
#endif


// Register write handler prototypes
static void RADSR(int, uint8);
static void REDl(int, uint8);
static void REFB(int, uint8);
static void REFBM(int, uint8);
static void REVolL(int, uint8);
static void REVolLM(int, uint8);
static void REVolR(int, uint8);
static void REVolRM(int, uint8);
static void RNOn(int, uint8);
static void RFCI(int, uint8);
static void RFlg(int, uint8);
static void RESA(int, uint8);
static void RGain(int, uint8);
static void RKOn(int, uint8);
static void RKOf(int, uint8);
static void RMVolL(int, uint8);
static void RMVolLM(int, uint8);
static void RMVolR(int, uint8);
static void RMVolRM(int, uint8);
static void RNull(int, uint8);
static void RPMOn(int, uint8);
static void RPitch(int, uint8);
static void RVolL(int, uint8);
static void RVolLM(int, uint8);
static void RVolR(int, uint8);
static void RVolRM(int, uint8);
static void REndX(int, uint8);

// Jump table for DSP register writes (see DSPIn)
static void (*dspRegs[0x80])(int, uint8) = {
	RVolL,  RVolR,  RPitch, RPitch, RNull,  RADSR,  RADSR,  RGain,
	RNull,  RNull,  RNull,  RNull,  RMVolL, REFB,   RNull,  RFCI,
	RVolL,  RVolR,  RPitch, RPitch, RNull,  RADSR,  RADSR,  RGain,
	RNull,  RNull,  RNull,  RNull,  RMVolR, RNull,  RNull,  RFCI,
	RVolL,  RVolR,  RPitch, RPitch, RNull,  RADSR,  RADSR,  RGain,
	RNull,  RNull,  RNull,  RNull,  REVolL, RPMOn,  RNull,  RFCI,
	RVolL,  RVolR,  RPitch, RPitch, RNull,  RADSR,  RADSR,  RGain,
	RNull,  RNull,  RNull,  RNull,  REVolR, RNOn,   RNull,  RFCI,
	RVolL,  RVolR,  RPitch, RPitch, RNull,  RADSR,  RADSR,  RGain,
	RNull,  RNull,  RNull,  RNull,  RKOn,   RNull,  RNull,  RFCI,
	RVolL,  RVolR,  RPitch, RPitch, RNull,  RADSR,  RADSR,  RGain,
	RNull,  RNull,  RNull,  RNull,  RKOf,   RNull,  RNull,  RFCI,
	RVolL,  RVolR,  RPitch, RPitch, RNull,  RADSR,  RADSR,  RGain,
	RNull,  RNull,  RNull,  RNull,  RFlg,   RESA,   RNull,  RFCI,
	RVolL,  RVolR,  RPitch, RPitch, RNull,  RADSR,  RADSR,  RGain,
	RNull,  RNull,  RNull,  RNull,  REndX,  REDl,   RNull,  RFCI
};

// Registers that require special handling for output type
static void (*dspRegsI[8])(int, uint8)= {  // Integer
    RVolL, RVolR, RMVolL, RMVolR, REVolL, REVolR, REFB, RFCI
};
static void (*dspRegsF[8])(int, uint8)= {  // Float
    RVolL, RVolR, RMVolL, RMVolR, REVolL, REVolR, REFB, RFCI
};
static void (*dspRegsM[8])(int, uint8)= {  // Monaural
    RVolLM, RVolRM, RMVolLM, RMVolRM, REVolLM, REVolRM, REFBM, RFCI
};

// Frequency table -------------------------
static uint32 freqTab[32] = {
	   0,
	2048, 1536, 1280,// Number of samples between updates.  Used to determine
	1024,  768,  640,// envelope rates and noise frequencies
	 512,  384,  320,
	 256,  192,  160,
	 128,   96,   80,
	  64,   48,   40,
	  32,   24,   20,
	  16,   12,   10,
	   8,    6,    5,
	   4,    3,
	   2,
	   1
};

// ADSR mode
static uint32 AttackRate [16] = {
	4100, 2600, 1500, 1000, 640, 380, 260, 160,
	  96,   64,   40,   24,  16,  10,   6,   1};

static uint32 DecayRate [8] = {
	1200, 740, 440, 290, 180, 110, 74, 37};

static uint32 SustainRate [32] = {
	  ~0, 38000, 28000, 24000, 19000, 14000, 12000, 9400,
	7100,  5900,  4700,  3500,  2900,  2400,  1800, 1500,
	1200,   880,   740,   590,   440,   370,   290,  220,
	 180,   150,   110,    92,    74,    55,    37,   18};

static uint32 IncreaseRate [32] = {
	 ~0, 4100, 3100, 2600, 2000, 1500, 1300, 1000,
	770,  640,  510,  380,  320,  260,  190,  160,
	130,   96,   80,   64,   48,   40,   32,   24,
	 20,   16,   12,   10,    8,    6,    4,    2};

static uint32 DecreaseRateExp [32] = {
	  ~0, 38000, 28000, 24000, 19000, 14000, 12000, 9400,
	7100,  5900,  4700,  3500,  2900,  2400,  1800, 1500,
	1200,   880,   740,   590,   440,   370,   290,  220,
	 180,   150,   110,    92,    74,    55,    37,   18};

// Floating point constants ----------------
static f32 fn2_5 = -2.5;						// Cubic interpolation
static f32 fn1_5 = -1.5;
static f32 fn0_5 = -0.5;
static f32 fp0_5 = 0.5;
static f32 fp1_5 = 1.5;

static f32 fp0_46 = 0.46;						// Low-pass filter
static f32 fp0_54 = 0.54;

static f32 fp32km1 = 32767.0;

static f32 fp512 = 512.0;						// Number of points of interpolation between samples
static f32 fp256 = 256.0;						// Number of points of interpolation between samples

static f64 fp32kHz =  32000.0;                  // SNES sample rate
static f64 fp32k   =  32767.0;                  // 2^15-1 (max 16-bit signed val)
static f64 fp4M    = 4294967296.0;              // 2**32

static uint64 max48;

// Prototypes for functions in mixRout
static bool8 EmuAPUDSPN(uint8 *&, int32 &);
static bool8 EmuAPUDSPI(uint8 *&, int32 &);
static bool8 EmuAPUDSPX(uint8 *&, int32 &);
static bool8 EmuAPUDSPF(uint8 *&, int32 &);

// Pointers to each mixing routine:  none, integer (386), mmx, and float
static bool8 (*mixRout[])(uint8 *&, int32 &) = {
	EmuAPUDSPN, EmuAPUDSPI, EmuAPUDSPX, EmuAPUDSPF
};

// Mixing ----------------------------------
static uint8 *pAPURAM;             // -> APU RAM
static uint8 *end_pAPURAM;         // -> APU RAM + 0xffff

static uint8  voiceKon;            // Voices that are currently being key on, == APU.KeyedChannels

VoiceMix mix[8];                   // Mixing settings for each voice and waveform playback

static uint32 rateTab[32];         // Update Rate Table

// DSP Options ----------------------
static int8   dspMix;              // Mixing routine
static int8   dspChn;              // Number of channels being output
static int8   dspSize;             // Size of samples in bytes
static int32  dspRate;             // Output sample rate
static int32  realRate;            // Real output sample rate, not emulated rate
static uint32 pitchAdj;            // Amount to adjust pitch rates [16.16]
static uint8  dspOpts;             // Option flags passed to SetDSPOpt

// Prototypes for functions that will go in dspDecmp
static inline void UnpckSrc(uint8, void *&, int16 *, int32 &, int32 &);

// Volume ----------------------------------
static int32  volAdj;              // Amount to adjust main volumes [-15.16]
static int32  volMainL;            // Main volume
static int32  volMainR;
static int32  volEchoL;            // Echo volume
static int32  volEchoR;

// Echo ------------------------------------
static uint32 echoStart;           // Echo Start Address
static uint32 echoDel;             // Size of delay (in bytes)
static uint32 echoCur;             // Current sample in echo area
static int32  echoFB;              // Echo feedback

// Noise -----------------------------------
static uint32 nRate;               // Noise sample rate reciprocal [.32]
static uint32 nAcc;                // Noise accumulator [.32] (>= 1 generate a new sample)
static int32  nSmp;                // Current Noise sample

// 64kHz counter ---------------------------
static uint32 t64Rate;
static uint32 t64Dec;

// Echo filtering --------------------------
static uint32 firCur;              // Index of the first sample to feed into the filter
static uint32 firRate;             // Rate to feed samples into filter
static uint8  firEnabl;            // 0 if filtering is disabled (each bit C0-C7 != 0)
static uint8  disEcho;             // 0 if echo is enabled

// Storage buffers -------------------------
extern int32 MixBuffer [SOUND_BUFFER_SIZE];
extern int32 EchoBuffer [SOUND_BUFFER_SIZE];
extern int32 DummyEchoBuffer [SOUND_BUFFER_SIZE];
extern int32 Echo [ECHOBUF];
extern int32 Loop [FIRBUF];
extern int32 FilterTaps [8];

// Other
uint32 t64Cnt;                     // 64kHz counter (increased every 64kHz pulse) by CPU Cycles
uint32 t64CntS;                    // 64kHz counter (increased every 64kHz pulse) by sample count
static uint32 konW;                // Last time the KON register was written to
static uint32 kofW;                // Last time the KOF register was written to
static SSoundData *sd;
static Channel *ch[NUM_CHANNELS];

//Used for converting envelope mode flags to Snes9x's enum
static int SAtoS9xEMode[256];

/***********************************************************************
  Code
***********************************************************************/

static inline uint16 DSP_GET_PITCH(int i)
{
	i <<= 4;
	return ((APU.DSP[i + APU_P_HIGH] << 8) |
			APU.DSP[i + APU_P_LOW]) & FREQUENCY_MASK;
}

static inline uint16 DSP_GET_SRC(int i)
{
	return READ_WORD(pAPURAM +
					 (((APU.DSP[APU_DIR] << 8) +
					   (APU.DSP[(i << 4) + APU_SRCN] << 2)) & 0xffff));
}

static inline uint16 DSP_GET_SRC_LOOP(int i)
{
	return READ_WORD(pAPURAM +
					 (((APU.DSP[APU_DIR] << 8) +
					   (APU.DSP[(i << 4) + APU_SRCN] << 2) + 2) & 0xffff));
}

// Start Waveform Decompression
//  Called when a voice is keyed on to set up the internal data for waveform
//  mixing, and decompress the first block.
static inline void StartSrc(int i, uint8 chm)
{
	// Initialize interpolation
	memset(&mix[i].sBuf[0x10], 0, sizeof(int16) * 16);

	mix[i].bCur = mix[i].bStart;
	mix[i].bHdr = *(uint8 *)mix[i].bCur;
	mix[i].sIdx = 0;
	mix[i].sP1 = 0;
	mix[i].sP2 = 0;

	// Start voice
	APU.DSP[APU_ENDX] &= ~chm;

	UnpckSrc(mix[i].bHdr,
			 mix[i].bCur,
			 mix[i].sBuf,
			 mix[i].sP1,
			 mix[i].sP2);
}

static inline void UpdateSrc(int i, uint8 chm)
{
	if(mix[i].bHdr & 1) {// Is this the end block?
		APU.DSP[APU_ENDX] |= chm;
		if(mix[i].bHdr & 2) {// Loop?
			mix[i].bCur = pAPURAM + DSP_GET_SRC_LOOP(i);
		}
		else {
			mix[i].mFlg |= MFLG_END;
			mix[i].mOut = 0;
			memset(&mix[i].sBuf[mix[i].sIdx & 0x10], 0, sizeof(int16) * 16);
			return;
		}
	}

	// Save block header
	mix[i].bHdr = *(uint8 *)mix[i].bCur;

	UnpckSrc(mix[i].bHdr,
			 mix[i].bCur,
			 &mix[i].sBuf[mix[i].sIdx & 0x10],
			 mix[i].sP1,
			 mix[i].sP2);
}

static inline void MixingFlagProcess(int i, uint8 chm)
{
	mix[i].mFlg &= ~MFLG_CHG;
	if(mix[i].mFlg & MFLG_SSRC) {
		mix[i].mFlg &= ~MFLG_SSRC;
		StartSrc(i, chm);
		mix[i].mFlg &= ~MFLG_END;
	}
}

//              Table 6.2.2. Noise Generator Clock
//
//     ----------------------------------------------------------------
//     | NCK | Freq   | NCK |   Freq   | NCK |   Freq  | NCK |   Freq  |
//     ----------------------------------------------------------------
//     |  00 |  0 Hz  |  08 |   83     |  10 |  500Hz  |  18 |   3.2   |
//     ----------------------------------------------------------------
//     |  01 |  16    |  09 |  100     |  11 |  667    |  19 |   4.0   |
//     ----------------------------------------------------------------
//     |  02 |  21    |  0A |  125     |  12 |  800    |  1A |   5.3   |
//     ----------------------------------------------------------------
//     |  03 |  25    |  0B |  107     |  13 |  1.0khz |  1B |   6.4   |
//     ----------------------------------------------------------------
//     |  04 |  31    |  0C |  200     |  14 |  1.3    |  1C |   8.0   |
//     ----------------------------------------------------------------
//     |  05 |  42    |  0D |  250     |  15 |  1.6    |  1D |  10.7   |
//     ----------------------------------------------------------------
//     |  06 |  50    |  0E |  333     |  16 |  2.0    |  1E |  16     |
//     ----------------------------------------------------------------
//     |  07 |  63    |  0F |  400     |  17 |  2.7    |  1F |  32     |
//     ----------------------------------------------------------------
static inline void SetNoiseHertz()
{
	int i = APU.DSP[APU_FLG] & 0x1f;
	if(i > 0) {
		nRate =  (uint32)(max48 / ((uint64)(freqTab[i] << 16) * dspRate / SNES_SAMPLE_RATE));
	}
	else {
		nRate = 0;
	}
}

//#define APU_VOL_LEFT 0x00
//#define APU_VOL_RIGHT 0x01
//#define APU_P_LOW 0x02
//#define APU_P_HIGH 0x03
//#define APU_SRCN 0x04
//#define APU_ADSR1 0x05
//#define APU_ADSR2 0x06
//#define APU_GAIN 0x07
static inline void InitReg(int reg, uint8 val)
{
	APU.DSP[reg] = val;
	int i = ((reg) & 0x70) >> 4;
	/* if(~reg & 0x08) the register is in dsp.voice */
	if(!((~(reg)) & MFLG_OFF & mix[i].mFlg))
		(*dspRegs[reg])(i, val);
}

void InitAPUDSP()
{
	int i, c;

	//max48
	max48 = 0x80000000;
	max48 <<= 17;
	max48 -= 1;

	pAPURAM = IAPU.RAM;// Save pointer to SPC RAM
	end_pAPURAM = pAPURAM + 0xffff;

	sd = &SoundData;
	for(i = 0; i < NUM_CHANNELS; i++) {
		ch[i] = &sd->channels[i];
	}

	// Reset values so SetDSPOpt will create new ones
	dspMix = -1;
	dspChn = -1;
	dspSize = -1;
    dspRate = -1;
	realRate = SNES_SAMPLE_RATE;
	dspOpts = 0;
    disEcho = 0;

     // Erase all mixer settings
    for(i = 0; i < 8; i++)
		memset(&mix[i], 0, sizeof(VoiceMix));

    // Set pointers to reasonable default values. Otherwise savestates cause
    // SEGVs!
    for(i = 0; i < 8; i++) {
		mix[i].sIdx = 0;
		mix[i].bStart = pAPURAM + DSP_GET_SRC(i);
		mix[i].bCur = pAPURAM + DSP_GET_SRC(i);
    }

	// Build a look-up table for all possible expanded values in a BRR block.
	// range 0-12
    for(i = 0; i < 13; i++) {
		for(c = 0; c < 16; c++)
			brrTab[(i << 4) + c] = (int32)(((((c ^ 8) - 8) << i) >> 1) << 1);
	}
	// range 13-15
    for(i = 13; i < 16; i++) {
		for(c = 0; c < 8; c++)
			brrTab[(i << 4) + c] = 0;
		for(c = 8; c < 16; c++)
			brrTab[(i << 4) + c] = -4096;
	}

	//Build a look-up table to calculate a cubic spline with only four integer multiplies.
	//The table is built from the following equation, simplified for s:
	//
	// y = ax^3 + bx^2 + cx + d
	//
	//     3 (s[0] - s[1]) - s[-1] + s[2]
	// a = ------------------------------
	//                   2
	//
	//                      5 s[0] + s[2]
	// b = 2 s[1] + s[-1] - -------------
	//                            2
	//
	//     s[1] - s[-1]
	// c = ------------
	//          2
	//
	// d = s[0]
	//
	//y is the return sample
	//x is the delta from current sample
	//s is a four sample array with [0] being the current sample
	f80 x1, x2, x3;
	for(i = 0; i < 256; i++) {
		//x1=(n/256)  x2=(n/256)^2  x3=(n/256)^3
		x1 = (f80)i / (f80)fp256;
		x2 = x1 * x1;
		x3 = x2 * x1;

		//s[-1] *= -.5(x^3) + (x^2) - .5x ------
		cubicTab[(i << 2) + 0] = (int16)(((f80)fn0_5 * x1 + x2 + (f80)fn0_5 * x3) * fp32km1);

		//s[0] *= 1.5(x^3) - 2.5(x^2) + 1 ------
		cubicTab[(i << 2) + 1] = (int16)(((f80)fn2_5 * x2 + (f80)fp1_5 * x3 + 1.0) * fp32km1);

		//s[1] *= -1.5(x^3) + 2(x^2) + .5x -----
		cubicTab[(i << 2) + 2] = (int16)(((f80)fp0_5 * x1 + x2 + x2 + (f80)fn1_5 * x3) * fp32km1);

		//s[2] *= .5(x^3) - .5(x^2) ------------
		cubicTab[(i << 2) + 3] = (int16)(((f80)fn0_5 * x2 + (f80)fp0_5 * x3) * fp32km1);
	}

#ifndef SA_SNESINT
	//Build a look-up table based on the Gaussian curve.
	//The table is built from the following equation:
	//
	//                  2
	//            (x/pi)
	//          - -------
	// y = A * e     2
	//
	//       ____
	//     \| 2pi * o
	// A = ----------
	//         n
	//
	//o is the mean value (2^15)
	//n is the number of points to interpolate between (4)
	f80 _512pi = (f80)fp512 / M_PIl;
	f32 fpA = (f32)((sqrt(M_PIl * 2) * 32768) / 4);

    for(i = 0; i < 256; i++) {
		f80 x;
		x = (f80)((-512 + i) + 0)   / _512pi;
		gaussTab[(i << 2) + 3] = (int16)(exp(-(x * x * 0.5)) * fpA);

		x = (f80)((-512 + i) + 256) / _512pi;
		gaussTab[(i << 2) + 2] = (int16)(exp(-(x * x * 0.5)) * fpA);

		x = (f80)((-512 + i) + 512) / _512pi;
		gaussTab[(i << 2) + 1] = (int16)(exp(-(x * x * 0.5)) * fpA);

		x = (f80)((-512 + i) + 768) / _512pi;
		gaussTab[(i << 2) + 0] = (int16)(exp(-(x * x * 0.5)) * fpA);
	}
#else
	//Interleave Gaussian table ---------------
    for(i = 0; i < 256; i++) {
		gaussTab[(i << 2) + 3] = gaussPre[i + 0];
		gaussTab[(i << 2) + 2] = gaussPre[i + 256];
		gaussTab[(i << 2) + 1] = gaussPre[i + 512];
		gaussTab[(i << 2) + 0] = gaussPre[i + 768];
	}
#endif

	// Create SAtoS9xEMode table
    for(i = 0; i < 256; i++)
		SAtoS9xEMode[i] = SOUND_SILENT;
	SAtoS9xEMode[E_IDLE] = SOUND_SILENT;
	SAtoS9xEMode[E_ATT] = SOUND_ATTACK;
	SAtoS9xEMode[E_DECAY] = SOUND_DECAY;
	SAtoS9xEMode[E_SUST] = SOUND_SUSTAIN;
	SAtoS9xEMode[E_REL] = SOUND_RELEASE;
	SAtoS9xEMode[E_DIRECT] = SOUND_GAIN;
	SAtoS9xEMode[E_INC] = SOUND_INCREASE_LINEAR;
	SAtoS9xEMode[E_BENT] = SOUND_INCREASE_BENT_LINE;
	SAtoS9xEMode[E_DEC] = SOUND_DECREASE_LINEAR;
	SAtoS9xEMode[E_EXP] = SOUND_DECREASE_EXPONENTIAL;

    SetAPUDSPOpt(MIX_NONE,
				 2,
				 16,
				 SNES_SAMPLE_RATE,
				 INT_GAUSS,
				 0);
}

void SetAPUDSPOpt(int mixType,
				  int numChn,
				  int bits,
				  int32 rate,
				  int inter,
				  int opts)
{
	int i, c;
	bool8 fixVol = FALSE;
	bool8 eraseBuf = FALSE;

	//=========================================
	// Verify parameters

	// mixType
	if(mixType == -1)
		mixType = dspMix;

	// numChn
	if(numChn == -1)
		numChn = dspChn;
	else
		numChn = numChn != 1 ? 2 : 1;

	// bits
	if(bits == -1)
		bits = dspSize;
	else {
		switch(bits) {
		case   8: break;
		case  16: break;
		case  24: break;
		case  32: break;
		case -32: break;
		default: bits = 16; break;
		}
		bits >>= 3;
	}

	// rate
	if(rate == -1)
		rate = (dspOpts & DSP_ANALOG) ? realRate : dspRate;
	else
		rate = (rate >= 8000 && rate <= 192000) ? rate : SNES_SAMPLE_RATE;

	// opts
	if(opts == -1)
		opts = dspOpts;

	// Verify mix type is capable of producing requested output
	if(mixType != 0 &&// MIX_NONE is unaffected by output settings
	   !(bits >= 2 && numChn >= 2 && mixType > MIX_INT)) {
		
		if(bits == 1) {
			mixType = MIX_INT;
		}
		else if(numChn == 1) {//numChn == mono
			mixType = MIX_INT;
			bits = 2;
		}
		else if(bits != 2) {
			mixType = MIX_FLOAT;
		}
	}

	// Reverse stereo
	if((dspOpts & DSP_REVERSE) ^ (opts & DSP_REVERSE))// Did reverse flag change?
		fixVol = TRUE;

	// Disable echo ----------------------------
	disEcho = (disEcho & ~DSP_NOECHO) | (opts & DSP_NOECHO);

	// Save option flags
	dspOpts = (uint8)opts;

	SetAPUDSPAmp(0x10000 * Options.VolumeLevel);

	//=========================================
	// Calculate sample rate change
	realRate = rate;
	dspOpts = rate < SNES_SAMPLE_RATE ? (dspOpts & ~DSP_ANALOG) : dspOpts;

	// Has sample rate changed?
	if(rate != dspRate) {
		dspRate = rate;

		// MAX is 0x400000
		pitchAdj = (uint32)(((uint64)SNES_SAMPLE_RATE << 20) / dspRate);

		//Calculate update rate for envelopes and noise
		uint64 r = (uint64)0x1000 * pitchAdj;
		rateTab[0] = 0;
		for(i = 1; i < 32; i++) {
			uint64 rr = r / freqTab[i];
			rateTab[i] = (uint32)((rr >> 16) + ((rr & 0xffff) ? 1 : 0));
		}

		// Noise
		SetNoiseHertz();

		//Sample counter
		r = (uint64)0x2000 * pitchAdj;
		t64Rate = (uint32)((r >> 16) + ((r & 0xffff) ? 1 : 0));
		t64Dec &= ~0xffff;

		firCur = 0;
		firRate = (uint32)(((int64)dspRate << 16) / SNES_SAMPLE_RATE);

		// Adjust voice rates
		for(i = 0; i < 7; i++) {
			mix[i].mOrgP = DSP_GET_PITCH(i);
			uint64 r = (uint64)mix[i].mOrgP * pitchAdj;
			mix[i].mRate = (uint32)((r >> 16) + ((r & 0xffff) ? 1 : 0));
			mix[i].mDec &= ~0xffff;
			mix[i].eRate = rateTab[mix[i].eRIdx];
			mix[i].eDec &= ~0xffff;
		}

		// Adjust echo delay
		InitReg(APU_ESA, APU.DSP[APU_ESA]);
		InitReg(APU_EDL, APU.DSP[APU_EDL]);

		eraseBuf = TRUE;
	}

	//=========================================
	// Set sample size
	if(bits != dspSize)
		dspSize = bits;

	//=========================================
	// Set number of channels
	if(numChn != dspChn) {
		dspChn = numChn;
		fixVol = TRUE;
	}

	//=========================================
	// Update areas affected by the mix type
	if(mixType != dspMix) {
		dspMix = mixType;
		fixVol = TRUE;
		eraseBuf = TRUE;
	}

	//=========================================
	// Erase sample buffers
	if(eraseBuf) {
		// Erase sample buffers
		memset(Echo, 0, sizeof(int32) * ECHOBUF);
		memset(Loop, 0, sizeof(int32) * FIRBUF);
	}

	//=========================================
	// Fixup volume handlers
	if(fixVol) {
		// Setup registers
		void (**regs)(int, uint8);
		regs = dspChn == 1 ? dspRegsM : (dspMix == MIX_FLOAT ? dspRegsF : dspRegsI);

		if(!(dspOpts & DSP_REVERSE)) {
			for(i = 0; i < 8; i++) {
				c = (i << 4);
				dspRegs[c + APU_VOL_LEFT] = regs[0];  // VOLL handler
				dspRegs[c + APU_VOL_RIGHT] = regs[1];  // VOLR handler
			}
			dspRegs[APU_MVOL_LEFT] = regs[2];   // MVOLL handler
			dspRegs[APU_MVOL_RIGHT] = regs[3];   // MVOLR handler
			dspRegs[APU_EVOL_LEFT] = regs[4];   // EVOLL handler
			dspRegs[APU_EVOL_RIGHT] = regs[5];   // EVOLR handler
		} else {
			// Reverse, e.g. put VOLL in VOLR slot
			for(i = 0; i < 8; i++) {
				c = (i << 4);
				dspRegs[c + APU_VOL_LEFT] = regs[1];  // VOLR handler
				dspRegs[c + APU_VOL_RIGHT] = regs[0];  // VOLL handler
			}
			dspRegs[APU_MVOL_LEFT] = regs[3];   // MVOLR handler
			dspRegs[APU_MVOL_RIGHT] = regs[2];   // MVOLL handler
			dspRegs[APU_EVOL_LEFT] = regs[5];   // EVOLR handler
			dspRegs[APU_EVOL_RIGHT] = regs[4];   // EVOLL handler
		}
		for(i = 0; i < 8; i++)
			dspRegs[(i << 4) + APU_C0] = regs[7];         // FC handler
		dspRegs[APU_EFB] = regs[6];  // EFB handler

		// Fix volumes and filter
		for(i = 0; i < 8; i++) {
			c = (i << 4);
			InitReg(c + APU_VOL_LEFT, APU.DSP[c + APU_VOL_LEFT]);
			InitReg(c + APU_VOL_RIGHT, APU.DSP[c + APU_VOL_RIGHT]);
			InitReg(c + APU_C0, APU.DSP[c + APU_C0]);
		}
		InitReg(APU_MVOL_LEFT, APU.DSP[APU_MVOL_LEFT]);
		InitReg(APU_MVOL_RIGHT, APU.DSP[APU_MVOL_RIGHT]);
		InitReg(APU_EVOL_LEFT, APU.DSP[APU_EVOL_LEFT]);
		InitReg(APU_EVOL_RIGHT, APU.DSP[APU_EVOL_RIGHT]);

		// Fix feedback level
		InitReg(APU_EFB, APU.DSP[APU_EFB]);
	}
}

void ResetAPUDSP(void)
{
	int i, c;

	// Erase DSP Registers ---------------------
	memset(APU.DSP, 0, 0x80);

	// Place DSP in power up mode
	APU.DSP[APU_FLG] = APU_SOFT_RESET | APU_MUTE;

	// Disable echo
	disEcho &= DSP_NOECHO;

    // Disable echo if all taps are 0
	disEcho |= 1;

    // Erase internal mixing settings
    // We'll just save-and-restore mFlg instead of selectively zeroing
	for(i = 0; i < 8; i++) {
		c = i << 4;

		// Voice is inactive
		mix[i].mFlg |= MFLG_OFF;
		mix[i].mFlg &= (MFLG_MUTE | MFLG_OFF);

		// Set pointers to reasonable default values. Otherwise savestates
		// cause SEGVs!
		mix[i].bStart = pAPURAM + DSP_GET_SRC(i);
		mix[i].bCur = pAPURAM + DSP_GET_SRC(i);
		mix[i].bHdr	= *(uint8 *)mix[i].bCur;

		// Envelope
		mix[i].eMode = E_IDLE;
		mix[i].eRIdx = 0;
		mix[i].eRate = rateTab[mix[i].eRIdx];
		mix[i].eDec &= ~0xffff;
		mix[i].eVal = 0;
		mix[i].eAdj = A_KOF;
		mix[i].eDest = D_MIN;
		
		// Samples
		memset(mix[i].sBuf, 0, sizeof(int16) * 32);
        mix[i].sIdx = 0;
		mix[i].sP1 = 0;
		mix[i].sP2 = 0;

		// Mixing
		mix[i].mChnL = 0;
		mix[i].mChnR = 0;
		RPitch(i, APU.DSP[c + APU_P_LOW]);
		mix[i].mDec &= ~0xffff;
		mix[i].mOut = 0;

 		// Previos sample data
		mix[i].pSampleL = 0;
		mix[i].pSampleR = 0;
	}

    // Erase global volume settings
    volMainL = 0;
    volMainR = 0;
    volEchoL = 0;
    volEchoR = 0;

    // Erase noise settings
    nRate = 0;
    nSmp = 0;
    nAcc = 0;

    // Erase echo region
    memset(Echo, 0, sizeof(int32) * ECHOBUF);

	// Reset echo variables
	echoStart = 0;
	echoCur = 0;
	echoDel = 2;  // Delay 1 sample
	echoFB = 0;

	// Echo filter
    memset(Loop, 0, sizeof(int32) * FIRBUF);// Erase filter memory
    memset(FilterTaps, 0, sizeof(int32) * 8);// Reset filter coefficients

	// Reset filter variables
    firCur = 0;
	firEnabl = 0;

    // Disable voices
	voiceKon = 0;

	// Reset times
	konW = 0;
	kofW = 0;
}

extern int NoiseFreq [32];

void StoreAPUDSP()
{
	int i, c;
	uint8 chm;

	APU.KeyedChannels = voiceKon;

	sd->master_volume_left  = (short)(int8)APU.DSP[APU_MVOL_LEFT];
	sd->master_volume_right = (short)(int8)APU.DSP[APU_MVOL_RIGHT];
	sd->echo_volume_left    = (short)(int8)APU.DSP[APU_EVOL_LEFT];
	sd->echo_volume_right   = (short)(int8)APU.DSP[APU_EVOL_RIGHT];
	sd->echo_enable         = APU.DSP[APU_EON];
	sd->echo_feedback       = (int)(int8)APU.DSP[APU_EFB];

	sd->echo_buffer_size    = (512 * (APU.DSP[APU_EDL] & 0x0f) * SNES_SAMPLE_RATE) / SNES_SAMPLE_RATE;
	sd->echo_buffer_size <<= 1;

	sd->echo_ptr            = 0;

	sd->echo_write_enabled  = (APU.DSP[APU_FLG] & APU_ECHO_DISABLED) == 0;
	sd->echo_channel_enable = APU.DSP[APU_EON];
	sd->pitch_mod           = APU.DSP[APU_PMON];

	sd->dummy[0] = 0;
	sd->dummy[1] = 0;
	sd->dummy[2] = 0;

	//no_filter
	S9xSetFilterCoefficient(0, (int)(signed char) APU.DSP [APU_C0]);

	sd->master_volume [0] = sd->master_volume_left;
	sd->master_volume [1] = sd->master_volume_right;
	sd->echo_volume [0] = sd->echo_volume_left;
	sd->echo_volume [1] = sd->echo_volume_right;

	sd->noise_hertz = NoiseFreq [APU.DSP[APU_FLG] & 0x1f];

	for (i = 0; i < 8; i++) {
		c = (i << 4);
		chm = (uint8)(1 << i);

		if(mix[i].mFlg & MFLG_CHG)
			MixingFlagProcess(i, chm);

		if (mix[i].mFlg & MFLG_OFF) {// Channel is inactive
			ch[i]->state = SOUND_SILENT;
			ch[i]->mode  = MODE_NONE;
		}
		else {
			ch[i]->state = SAtoS9xEMode[mix[i].eMode];
			if (mix[i].mFlg & MFLG_KOFF)// Key off state
				ch[i]->mode = MODE_RELEASE;
			else {
				if (APU.DSP[c + APU_ADSR1] & 0x80)  // ADSR mode
					ch[i]->mode = MODE_ADSR;
				else {// GAIN mode
					if (!(APU.DSP[c + APU_GAIN] & 0x80))  // Direct mode
						ch[i]->mode = MODE_GAIN;
					else {                                       // Indirect mode
						switch(APU.DSP[c + APU_GAIN] & 0x60){
						case 0x00:
							ch[i]->mode = MODE_DECREASE_LINEAR;
							break;

						case 0x20:
							ch[i]->mode = MODE_DECREASE_EXPONENTIAL;
							break;

						case 0x40:
							ch[i]->mode = MODE_INCREASE_LINEAR;
							break;

						case 0x60:
							ch[i]->mode = MODE_INCREASE_BENT_LINE;
							break;
						}

					}
				}
            }
		}
			
		ch[i]->type            = (APU.DSP[APU_NON] & chm) ? SOUND_NOISE : SOUND_SAMPLE;
		ch[i]->volume_left     = (short)(int8)APU.DSP[c + APU_VOL_LEFT];
		ch[i]->volume_right    = (short)(int8)APU.DSP[c + APU_VOL_RIGHT];
		ch[i]->hertz           = ((uint32)DSP_GET_PITCH(i) * SNES_SAMPLE_RATE) >> 12;
		ch[i]->frequency       = (uint32)(((int64)(ch[i]->type == SOUND_NOISE ?
												   sd->noise_hertz : ch[i]->hertz) *
										   FIXED_POINT) / SNES_SAMPLE_RATE);

		ch[i]->count           = 0;
		ch[i]->envx            = mix[i].eVal >> E_SHIFT;
		ch[i]->envxx           = (ch[i]->envx << 24);// #define ENVX_SHIFT 24
		ch[i]->left_vol_level  = (ch[i]->envx * ch[i]->volume_left) / 128;
		ch[i]->right_vol_level = (ch[i]->envx * ch[i]->volume_right) / 128;

		switch(mix[i].eMode) {
		case E_ATT:
			//Attack mode
			//ch[i]->envx_target = MAX_ENVELOPE_HEIGHT;
			ch[i]->envx_target = mix[i].eDest >> E_SHIFT;
			break;

		case E_DECAY:
			//Decay mode
			//ch[i]->envx_target = (MAX_ENVELOPE_HEIGHT * ((APU.DSP[c + APU_ADSR2] >> 5) + 1)) >> 3;
			ch[i]->envx_target = mix[i].eDest >> E_SHIFT;
			break;

		case E_SUST:
			//Sustain mode
			//ch[i]->envx_target = MAX_ENVELOPE_HEIGHT / 10;
			ch[i]->envx_target = mix[i].eDest >> E_SHIFT;
			break;

		default:
			ch[i]->envx_target = 0;
			break;
		}

		ch[i]->env_error       = mix[i].eDec;
		if(so.playback_rate > 0)
			ch[i]->erate       = mix[i].eRate * SNES_SAMPLE_RATE / so.playback_rate;
		else
			ch[i]->erate = 0;

		ch[i]->direction       = mix[i].eRate == 0 ? 0 : ((mix[i].eMode & E_DIR) ? 1 : -1);
		ch[i]->attack_rate     = AttackRate[APU.DSP[c + APU_ADSR1] & 0xF];
		ch[i]->decay_rate      = DecayRate[(APU.DSP[c + APU_ADSR1] >> 4) & 7];
		ch[i]->sustain_rate    = SustainRate[APU.DSP[c + APU_ADSR2] & 0x1F];
		ch[i]->release_rate    = 8;
		ch[i]->sustain_level   = (APU.DSP[c + APU_ADSR2] >> 5) + 1;
		ch[i]->interpolate     = 0;

		memcpy(ch[i]->decoded, &mix[i].sBuf[mix[i].sIdx & 0x10], sizeof(int16) * 16);

		ch[i]->sample          = mix[i].sBuf[mix[i].sIdx];

		if(mix[i].sIdx < 31)
			ch[i]->next_sample = mix[i].sBuf[mix[i].sIdx + 1];
		else
			ch[i]->next_sample = 0;

		ch[i]->sample_pointer  = (uint32)(mix[i].sIdx & 0x0f);
		ch[i]->Filter          = mix[i].bHdr;
		ch[i]->last_block      = (ch[i]->Filter & 1) != 0;
		ch[i]->loop            = (ch[i]->Filter & 2) != 0;

		ch[i]->previous[0]   = mix[i].sP1;
		ch[i]->previous[1]   = mix[i].sP2;
		ch[i]->previous16[0]   = (signed short)mix[i].sP1;
		ch[i]->previous16[1]   = (signed short)mix[i].sP2;
		ch[i]->block = (signed short *)&ch[i]->decoded[ch[i]->sample_pointer];
		ch[i]->sample_number   = APU.DSP [c + APU_SRCN];
		ch[i]->needs_decode    = FALSE;
		ch[i]->block_pointer   = (uint32)((uint8 *)mix[i].bCur - pAPURAM);

		if (APU.DSP [APU_EON] & chm)
			ch[i]->echo_buf_ptr = EchoBuffer;
		else
			ch[i]->echo_buf_ptr = DummyEchoBuffer;

		ch[i]->chkMode = 0;

		ch[i]->dummy[0] = 0;
		ch[i]->dummy[1] = 0;
		ch[i]->dummy[2] = 0;
		ch[i]->dummy[3] = 0;
		ch[i]->dummy[4] = 0;
		ch[i]->dummy[5] = 0;
		ch[i]->dummy[6] = 0;
		ch[i]->dummy[7] = 0;

		// Store OUTX ENVX
		APU.DSP[c + APU_OUTX] = (uint8)((mix[i].mFlg & (MFLG_OFF | MFLG_END)) ? 0 : (mix[i].mOut >> 8));
		APU.DSP[c + APU_ENVX] = (uint8)((mix[i].eVal >> E_SHIFT) & 0x7f);
	}

	// Store 0xf3
	pAPURAM[0xf3] = APU.DSP[pAPURAM[0xf2] & 0x7f];
}

static const uint8 S9xtoSAEMode[10] = {
	E_IDLE,  // SOUND_SILENT
	E_ATT,   // SOUND_ATTACK
	E_DECAY, // SOUND_DECAY
	E_SUST,  // SOUND_SUSTAIN
	E_REL,   // SOUND_RELEASE
	E_DIRECT,// SOUND_GAIN
	E_INC,   // SOUND_INCREASE_LINEAR
	E_BENT,  // SOUND_INCREASE_BENT_LINE
	E_DEC,   // SOUND_DECREASE_LINEAR
	E_EXP    // SOUND_DECREASE_EXPONENTIAL
};

// Restore DSP's State of Operation
void RestoreAPUDSP()
{
	int i, c;

	voiceKon = APU.KeyedChannels;

	for (i = 0; i < 8; i++) {
		c = (i << 4);

		mix[i].bStart = pAPURAM + DSP_GET_SRC(i);
		mix[i].bCur	= pAPURAM + (ch[i]->block_pointer & 0xffff);
		mix[i].bHdr	= (uint8)ch[i]->Filter;
		memset(&mix[i].sBuf[0x10], 0, sizeof(int16) * 16);
		memcpy(mix[i].sBuf, ch[i]->decoded, sizeof(int16) * 16);
		mix[i].sP1	= (int32)ch[i]->previous16[0];
		mix[i].sP2	= (int32)ch[i]->previous16[1];
		mix[i].sIdx = (uint8)(ch[i]->sample_pointer & 0x0f);

		uint8 mFlg = ((mix[i].mFlg & MFLG_MUTE) |
					  ((ch[i]->last_block && !ch[i]->loop) ? MFLG_END : 0) |
					  (ch[i]->state == SOUND_SILENT ? MFLG_OFF : 0) |
					  (ch[i]->mode == MODE_RELEASE ? MFLG_KOFF : 0));

		if(mFlg & MFLG_OFF)
			mFlg &= ~(MFLG_KOFF | MFLG_END);

		mix[i].mFlg	= mFlg;

		InitReg(c + APU_VOL_LEFT, APU.DSP[c + APU_VOL_LEFT]);
		InitReg(c + APU_VOL_RIGHT, APU.DSP[c + APU_VOL_RIGHT]);
		InitReg(c + APU_P_LOW, APU.DSP[c + APU_P_LOW]);
		
		mix[i].eMode = S9xtoSAEMode[ch[i]->state];

		// Default to release mode settings
		mix[i].eRIdx = 31;
		mix[i].eRate = rateTab[mix[i].eRIdx];
		mix[i].eDec &= ~0xffff;
		mix[i].eAdj = A_KOF;
		mix[i].eDest = D_MIN;
		mix[i].eVal = ch[i]->envx << E_SHIFT;
		InitReg(c + APU_ADSR1, APU.DSP[c + APU_ADSR1]);

		mix[i].mOut = (int32)(int8)APU.DSP[c + APU_OUTX] << 8;
		mix[i].mDec &= ~0xffff;
	}
    // Setup global paramaters
    InitReg(APU_MVOL_LEFT, APU.DSP[APU_MVOL_LEFT]);
    InitReg(APU_MVOL_RIGHT, APU.DSP[APU_MVOL_RIGHT]);
    InitReg(APU_EVOL_LEFT, APU.DSP[APU_EVOL_LEFT]);
    InitReg(APU_EVOL_RIGHT, APU.DSP[APU_EVOL_RIGHT]);

	// Don't "APU_SOFT_RESET"
	uint8 flg = APU.DSP[APU_FLG];
	RFlg((APU_FLG & 0x70) >> 4, flg & ~APU_SOFT_RESET);
	APU.DSP[APU_FLG] = flg;

    InitReg(APU_ESA, APU.DSP[APU_ESA]);
    InitReg(APU_EFB, APU.DSP[APU_EFB]);
    InitReg(APU_EDL, APU.DSP[APU_EDL]);

    for(i = 0; i < 8; i++) {
		c = (i << 4);
		InitReg(c + APU_C0, APU.DSP[c + APU_C0]);
    }
}

void RestoreSPCFile()
{
	int i, c;

	for (i = 0; i < 8; i++) {
		c = (i << 4);

		mix[i].eVal = APU.DSP[c + APU_ENVX] << E_SHIFT;
		//mix[i].mOut = (int32)(int8)APU.DSP[c + APU_OUTX] << 8;
	}

    // Setup global paramaters
    InitReg(APU_MVOL_LEFT, APU.DSP[APU_MVOL_LEFT]);
    InitReg(APU_MVOL_RIGHT, APU.DSP[APU_MVOL_RIGHT]);
    InitReg(APU_EVOL_LEFT, APU.DSP[APU_EVOL_LEFT]);
    InitReg(APU_EVOL_RIGHT, APU.DSP[APU_EVOL_RIGHT]);

	// Don't "APU_SOFT_RESET"
	uint8 flg = APU.DSP[APU_FLG];
	RFlg((APU_FLG & 0x70) >> 4, flg & ~APU_SOFT_RESET);
	APU.DSP[APU_FLG] = flg;

    InitReg(APU_ESA, APU.DSP[APU_ESA]);
    InitReg(APU_EFB, APU.DSP[APU_EFB]);
    InitReg(APU_EDL, APU.DSP[APU_EDL]);

    for(i = 0; i < 8; i++) {
		c = (i << 4);
		InitReg(c + APU_C0, APU.DSP[c + APU_C0]);
    }

	// Start Voices
	InitReg(APU_KON, APU.DSP[APU_KON]);
}

// DSP Pre-Amplification Level
void SetAPUDSPAmp(int32 amp)
{
	volAdj = (int32)(((int64)amp * 0x10000) >> 16);

	// Update global volumes
	InitReg(APU_MVOL_LEFT, APU.DSP[APU_MVOL_LEFT]);
	InitReg(APU_MVOL_RIGHT, APU.DSP[APU_MVOL_RIGHT]);
	InitReg(APU_EVOL_LEFT, APU.DSP[APU_EVOL_LEFT]);
	InitReg(APU_EVOL_RIGHT, APU.DSP[APU_EVOL_RIGHT]);
}

// Start Envelope
//  Called when a voice is keyed on to set up the internal data to begin
//  envelope modification based on the values in ADSR/Gain.
static inline void StartEnv(int idx)
{
    mix[idx].eDec &= ~0xffff;        // Clear decimal overflow

    if(APU.DSP[idx*0x10+APU_ADSR1]&0x80){  // Is the envelope in ADSR mode?
        mix[idx].eVal=D_MIN;
// ChgAtt starts here, see ChgEnv()
        if((APU.DSP[idx*0x10+APU_ADSR1]&0x0f)!=0x0f){ // Is there an attack?
            mix[idx].eRIdx=((APU.DSP[idx*0x10+APU_ADSR1]&0x0f)<<1)+1;
            mix[idx].eRate=rateTab[mix[idx].eRIdx];  // Rate of adjustment
            mix[idx].eAdj=A_LIN;      // Set adjustment rate to linear
            mix[idx].eDest=D_ATTACK;  // Set destination to 1.0
            mix[idx].eMode=E_ATT;      // Set envelope mode to attack
            return;
        }
        // No, Start with decay
        mix[idx].eVal=D_MAX;  // No attack, so envelope starts at 1.0
// ChgDec starts here, see ChgEnv()
        mix[idx].eAdj=A_EXP;  // Set adjustment rate to exponential
        if((APU.DSP[idx*0x10+APU_ADSR2]>>5)!=7){
            mix[idx].eDest=((APU.DSP[idx*0x10+APU_ADSR2]>>5)+1)*D_EXP;
            mix[idx].eRIdx=((APU.DSP[idx*0x10+APU_ADSR1]&0x70)>>3)+0x10;
            mix[idx].eRate=rateTab[mix[idx].eRIdx];  // Rate of adjustment
            mix[idx].eMode=E_DECAY;  // Set envelope mode to decay
            return;
        }
// ChgSus starts here, see ChgEnv()
        mix[idx].eRIdx=APU.DSP[idx*0x10+APU_ADSR2]&0x1f;
        mix[idx].eRate=rateTab[mix[idx].eRIdx];
        mix[idx].eDest=D_MIN;   // Set destination to 0.1
        mix[idx].eMode=E_SUST;        // Set envelope mode to sustain
        return;
    }
    // Not ADSR, It's in gain mode
    mix[idx].eMode = E_IDLE;    // Set envelope mode to constant
    if(!(APU.DSP[idx*0x10+APU_GAIN]&0x80)){  // Is gain direct?
          mix[idx].eVal=(APU.DSP[idx*0x10+APU_GAIN]&0x7f)*A_GAIN;
        return;
    }

    // No, Program envelope
    mix[idx].eRIdx=APU.DSP[idx*0x10+APU_GAIN]&0x1f;
    mix[idx].eRate=rateTab[mix[idx].eRIdx];
    if(mix[idx].eRate==0) return;
    switch(APU.DSP[idx*0x10+APU_GAIN]&0x60){
      case 0x00:
        // GainDec
        mix[idx].eAdj=A_LIN;
        mix[idx].eDest=D_MIN;
        mix[idx].eMode=E_DEC;       // Set mode to linear decrease
        break;

      case 0x20:
        // GainExp
        mix[idx].eAdj=A_EXP;
        mix[idx].eDest=D_MIN;
        mix[idx].eMode=E_EXP;       // Set mode to exponential decrease
        break;

      case 0x40:
        // GainInc
        mix[idx].eAdj=A_LIN;
        mix[idx].eDest=D_ATTACK;
        mix[idx].eMode=E_INC;       // Set mode to linear increase
        break;

      case 0x60:
        // GainBent
        mix[idx].eAdj=A_LIN;
        mix[idx].eDest=D_BENT;
        mix[idx].eMode=E_BENT;       // Set mode to bent line increase
        break;
    }
}

static inline void ChgEnv(int idx)
{
    // If the envelope isn't in attack, decay, or sustain mode, changes to the
    // ADSR registers have no effect
    if(mix[idx].eMode==E_ATT) goto ChgAtt;
    if(mix[idx].eMode==E_DECAY) goto ChgDec;
    if(mix[idx].eMode==E_SUST) goto ChgSus;
    return;

// ChgAtt starts here, see StartEnv()
ChgAtt:
    if((APU.DSP[idx*0x10+APU_ADSR1]&0x0f)!=0x0f){ // Is there an attack?
        mix[idx].eRIdx=((APU.DSP[idx*0x10+APU_ADSR1]&0x0f)<<1)+1;
        mix[idx].eRate=rateTab[mix[idx].eRIdx];  // Rate of adjustment
        mix[idx].eAdj=A_LIN;      // Set adjustment rate to linear
        mix[idx].eDest=D_ATTACK;  // Set destination to 1.0
        mix[idx].eMode=E_ATT;      // Set envelope mode to attack
        return;
    }
    // No, Start with decay
    mix[idx].eVal=D_MAX;  // No attack, so envelope starts at 1.0
// ChgDec starts here, see StartEnv()
ChgDec:
    mix[idx].eAdj=A_EXP;  // Set adjustment rate to exponential
    if((APU.DSP[idx*0x10+APU_ADSR2]>>5)!=7){
        mix[idx].eDest=((APU.DSP[idx*0x10+APU_ADSR2]>>5)+1)*D_EXP;
        mix[idx].eRIdx=((APU.DSP[idx*0x10+APU_ADSR1]&0x70)>>3)+0x10;
        mix[idx].eRate=rateTab[mix[idx].eRIdx];  // Rate of adjustment
        mix[idx].eMode=E_DECAY;  // Set envelope mode to decay
        return;
    }
// ChgSus starts here, see StartEnv()
ChgSus:
    mix[idx].eRIdx=APU.DSP[idx*0x10+APU_ADSR2]&0x1f;
    mix[idx].eRate=rateTab[mix[idx].eRIdx];
    mix[idx].eDest=D_MIN;   // Set destination to 0.1
    mix[idx].eMode=E_SUST;        // Set envelope mode to sustain
    return;
}

// DSP Data Port
void APUDSPIn(uint8 address, uint8 data)
{
	// Writes to 80-FFh have no effect (reads are mirrored from lower mem)
	if(address & 0x80)
		return;

	switch(address) {
	case APU_KON:
	case APU_KOFF:
		InitReg(address, data);
		break;

	case APU_FLG:
		RFlg((address & 0x70) >> 4, data);
		break;

	case APU_ENDX:
		REndX((address & 0x70) >> 4, data);
		break;

	default:
		if(data != APU.DSP[address])
			InitReg(address, data);
		break;
	}
}

/***********************************************************************
  DSP Register Handlers
***********************************************************************/

// Channel volume
static void RVolL(int i, uint8 val)
{
	mix[i].mChnL = (int32)(int8)val;
}

static void RVolLM(int i, uint8 val)
{
	mix[i].mChnL = (int32)ABS((int8)val);
}

static void RVolR(int i, uint8 val)
{
	mix[i].mChnR = (int32)(int8)val;
}

static void RVolRM(int i, uint8 val)
{
	mix[i].mChnR = (int32)ABS((int8)val);
}

// Pitch
static void RPitch(int i, uint8)
{
	mix[i].mOrgP = DSP_GET_PITCH(i);
	uint64 r = (uint64)mix[i].mOrgP * pitchAdj;
	mix[i].mRate = (uint32)((r >> 16) + ((r & 0xffff) ? 1 : 0));
}

// Envelope
static void RADSR(int i, uint8 val)
{
	if(mix[i].mFlg & MFLG_KOFF)
		return;
	if(!(APU.DSP[i*0x10+APU_ADSR1] & 0x80))
		RGain(i, val);
	else
		ChgEnv(i);
}

static void RGain(int i, uint8)
{
	if(mix[i].mFlg & MFLG_KOFF)
		return;
	if(APU.DSP[i*0x10+APU_ADSR1] & 0x80)
		return;
	StartEnv(i);
}

// Main volumes
static void RMVolL(int, uint8 val)
{
	if(Settings.DisableMasterVolume)
		val = 127;
    volMainL = ((int32)(int8)val * volAdj) >> 16;
}

static void RMVolLM(int, uint8 val)
{
	if(Settings.DisableMasterVolume)
		val = 127;
	// Multiply by 1/2 sqrt 2 to help average
	volMainL = (int32)(((int64)ABS((int8)val) * 46341 * volAdj) >> 32);
}

static void RMVolR(int, uint8 val)
{
	if(Settings.DisableMasterVolume)
		val = 127;
	volMainR = ((int32)(int8)val * volAdj) >> 16;
}

static void RMVolRM(int, uint8 val)
{
	if(Settings.DisableMasterVolume)
		val = 127;
	// Multiply by 1/2 sqrt 2 to help average
	volMainR = (int32)(((int64)ABS((int8)val) * 46341 * volAdj) >> 32);
}

static void REVolL(int, uint8 val)
{
	volEchoL = ((int32)(int8)val * volAdj) >> 16;
}

static void REVolLM(int, uint8 val)
{
	// Multiply by 1/2 sqrt 2 to help average
    volEchoL = (int32)(((int64)ABS((int8)val) * 46341 * volAdj) >> 32);
}

static void REVolR(int, uint8 val)
{
	volEchoR = ((int32)(int8)val * volAdj) >> 16;
}

static void REVolRM(int, uint8 val)
{
	// Multiply by 1/2 sqrt 2 to help average
	volEchoR = (int32)(((int64)ABS((int8)val) * 46341 * volAdj) >> 32);
}

// Echo settings
static void REFB(int, uint8 val)
{
	echoFB = (int32)(int8)val;
}

static void REFBM(int, uint8 val)
{
	echoFB = (int32)(int8)val;
}

static void REDl(int, uint8 val)
{
	val &= 0x0f;
	if(val == 0)
		echoDel = 2;
	else
		echoDel = ((uint32)(val << 4) * dspRate / 1000) << 1;
}

static inline void CheckFC(int i)
{
	// If first tap is 127, pretend it's 0
	if(!(i == 0 && APU.DSP[APU_C0] == 0x7f) && APU.DSP[(i << 4) + APU_C0])
		firEnabl |= (1 << i);
	else
		firEnabl &= ~(1 << i);

    // Disable echo if all taps are 0
	if(!(firEnabl | APU.DSP[APU_C0]))
		disEcho |= 1;
	else
		disEcho &= ~1;
}

static void RFCI(int i, uint8 val)
{
	FilterTaps[i] = (int32)(int8)val;
	CheckFC(i);
}

static inline void UpdateEnv(int i, uint8 chm)
{
    if(mix[i].eMode&E_TYPE){
        // Is the adjustment exponential? (mode 1, 9, or D)
        mix[i].eVal-=mix[i].eVal>>8;         // Subtract 1/256th of envelope height
        if(mix[i].eVal>mix[i].eDest) return; // Has height reached destination? No, quit.
        mix[i].eVal = mix[i].eDest;
    } else if(mix[i].eMode&E_DIR){
        // Linear increase (mode 2, 6, or A)
        mix[i].eVal+=mix[i].eAdj;            // Add adjustment to height
        if(mix[i].eVal<mix[i].eDest) return; // Has height reached destination? No, quit.
        mix[i].eVal = mix[i].eDest;
    } else {
        // Linear decrease (mode 0 or 8)
		if(mix[i].eVal > mix[i].eDest) {
			if(mix[i].eVal >= mix[i].eAdj) {
				mix[i].eVal -= mix[i].eAdj;   // Subtract adjustment from height
				if(mix[i].eVal > mix[i].eDest)
					return;   // Has height reached 0? No, quit.
			}
			mix[i].eVal = mix[i].eDest;
		}
        if(mix[i].eMode == E_REL) {  // Is voice in release mode?
            mix[i].mFlg |= MFLG_OFF; // Turn voice off
            mix[i].mFlg &= ~MFLG_KOFF; // Turn voice off
			mix[i].mDec &= ~0xffff;
			mix[i].eDec &= ~0xffff;
			mix[i].mOut = 0;
        }
        goto EnvConst;
    }
    // Change adjustment mode (see StartEnv)
    if(mix[i].eMode&E_ADSR){ // Is envelope in ADSR mode?
        switch(mix[i].eMode){
          case E_ATT:
            mix[i].eAdj=A_EXP;
            if((APU.DSP[i*0x10+APU_ADSR2]>>5)!=7){
                mix[i].eDest=((APU.DSP[i*0x10+APU_ADSR2]>>5)+1)*D_EXP;
                mix[i].eRIdx=((APU.DSP[i*0x10+APU_ADSR1]&0x70)>>3)+0x10;
                mix[i].eRate=rateTab[mix[i].eRIdx];
                mix[i].eMode=E_DECAY;
                return;
            }
            /* else fall through... */
          case E_DECAY:
            mix[i].eRIdx=APU.DSP[i*0x10+APU_ADSR2]&0x1f;
            mix[i].eRate=rateTab[mix[i].eRIdx];
            mix[i].eDec &= ~0xffff;
            mix[i].eAdj=A_EXP;
            mix[i].eDest=D_MIN;
            mix[i].eMode=E_SUST;
            return;
          default:
			mix[i].eRIdx=31;
			mix[i].eRate=rateTab[mix[i].eRIdx];// 1->0 8ms
            mix[i].eDec &= ~0xffff;
            mix[i].eAdj=A_KOF;  // Set adjustment rate to linear
            mix[i].eDest=D_MIN; // Set destination to 0
            mix[i].eMode=E_REL;     // Set envelope mode to release
            return;
        }
    } else {
        if(!(mix[i].eMode&E_DEST)) { // Is gain in bent line mode? No, it's constant
             goto EnvConst;
        }
        mix[i].eAdj=A_BENT;
        mix[i].eDest=D_MAX;  // Switch to bent mode
        mix[i].eMode=E_INC;
        return;
    }
EnvConst:
	mix[i].eMode = E_IDLE;
}

// Key On/Off
static void RKOn(int idx, uint8 val)
{
	konW = t64Cnt;

	if (spc_is_dumping)
	{
		if (val & ~spc_is_dumping_temp)
		{
			APURegisters.PC = IAPU.PC - IAPU.RAM;
			S9xAPUPackStatus();
#ifdef __UOSNES__
  #ifdef __WIN32__
			extern void S9xwSPCDumpMessages(bool8 err);
			S9xwSPCDumpMessages(S9xSPCDump (S9xGetSPCFilename ()));
  #else
			S9xSPCDump (S9xGetSPCFilename ());
  #endif
#else
			S9xSPCDump (S9xGetFilenameInc (".spc"));
#endif
			spc_is_dumping = 0;
		}
	}
	spc_is_dumping_temp = val;

	//if((t64Cnt - kofW) < 4)
	//	val &= ~APU.DSP[APU_KOFF];

	if(val) {
		uint8 chm = 1;
		for(int i = 0; i < 8; i++, chm <<= 1) {
			if(!(val & chm))
				continue;

			int c = (i << 4);

			// Set channel volume
			(*dspRegs[c + APU_VOL_LEFT])(i, APU.DSP[c + APU_VOL_LEFT]);
			(*dspRegs[c + APU_VOL_RIGHT])(i, APU.DSP[c + APU_VOL_RIGHT]);

			// Set pitch
			RPitch(i, APU.DSP[c + APU_P_LOW]);
			mix[i].mDec &= ~0xffff;

			// Start envelope
			StartEnv(i);

			// Start waveform decompression
			mix[i].bStart = pAPURAM + DSP_GET_SRC(i);
			mix[i].mFlg |= (MFLG_CHG | MFLG_SSRC);
			mix[i].mFlg &= ~(MFLG_KOFF | MFLG_OFF);// Reset flags

			// Mark voice as being on internally
			voiceKon |= chm;
		}
	}
}

static void RKOf(int, uint8 val)
{
	kofW = t64Cnt;

	if((t64Cnt - konW) < 4)
		val &= ~APU.DSP[APU_KON];

	// Only check voices that are currently playing
	val &= voiceKon;

	if(val) {
		uint8 chm = 1;
		for(int i = 0; i < 8; i++, chm <<= 1) {
			if(!(val & chm))
				continue;

			mix[i].eRIdx = 31;// 1->0 8ms
			mix[i].eRate = rateTab[mix[i].eRIdx];
			mix[i].eAdj = A_KOF;
			mix[i].eDest = D_MIN;
			mix[i].eMode = E_REL;

			mix[i].mFlg &= ~MFLG_SSRC;// Reset flags
			mix[i].mFlg |= MFLG_KOFF;// Flag voice as keying off

			voiceKon &= ~chm;
		}
	}
}

static void RPMOn(int, uint8)
{
	for(int i = 1; i < 8; i++) {
		if(APU.DSP[APU_PMON] & (1 << i))
			continue;
		//Reset pitch
		uint64 r = (uint64)mix[i].mOrgP * pitchAdj;
		mix[i].mRate = (uint32)((r >> 16) + ((r & 0xffff) ? 1 : 0));
	}
}

static void RNOn(int, uint8)
{
	for(int i = 1; i < 8; i++) {
		if(!(APU.DSP[APU_NON] & APU.DSP[APU_PMON] & (1 << i)))
			continue;
		//Reset pitch
		uint64 r = (uint64)mix[i].mOrgP * pitchAdj;
		mix[i].mRate = (uint32)((r >> 16) + ((r & 0xffff) ? 1 : 0));
	}
}

static void RFlg(int, uint8 al)
{
	if(al & APU_SOFT_RESET) {// Has a soft reset been initialized?
		al |= APU_MUTE;

		// Reset internal voice settings
		for(int i = 0; i < 8; i++) {
			mix[i].mFlg |= MFLG_OFF;
            mix[i].mFlg &= (MFLG_MUTE | MFLG_OFF);
			mix[i].mDec &= ~0xffff;
			mix[i].mOut = 0;
			mix[i].eMode = E_IDLE;
			mix[i].eDec &= ~0xffff;
		}

		APU.DSP[APU_ENDX] = 0;// Clear end block flags
		APU.DSP[APU_KON] = 0;
		APU.DSP[APU_KOFF] = 0;
		voiceKon = 0;
	}

	// Disable echo
	disEcho = (disEcho & ~APU_ECHO_DISABLED) | (al & APU_ECHO_DISABLED);
	APU.DSP[APU_FLG] = al;
	SetNoiseHertz();
}

static void RESA(int, uint8 val)
{
	echoStart = (uint32)val * 128 * dspRate / SNES_SAMPLE_RATE;
}

static void REndX(int, uint8)
{
	APU.DSP[APU_ENDX] = 0;
}

// Null register
static void RNull(int, uint8)
{
	/* Do nothing */
}

// Noise Generator
//  Generates white noise samples (into nSmp)
static inline void NoiseGen()
{
	nAcc += nRate;
	if(nAcc < nRate)
		nSmp = (int32)(int16)(nSmp * 27865 + 7263);
}

// Pitch Modulation
//
// Changes the pitch based on the output of the previous voice:
//
//  P' = (P * (OUTX + 32768)) >> 15
//
// Pitch modulation in the SNES uses the full 16-bit sample value, not the 8-bit value in OUTX as
// previously believed.
static inline void PitchMod(int i, uint8 chm)
{
	if(i > 0 &&
	   (APU.DSP[APU_PMON] & chm)) {
		
		int32 t = ((mix[i - 1].mOut + 32768) * mix[i].mOrgP) >> 15;

		if(t > 0x3FFF)
			t = 0x3FFF;
		else if(t < 0)
			t = 0;

		uint64 r = (uint64)t * pitchAdj;
		mix[i].mRate = (uint32)((r >> 16) + ((r & 0xffff) ? 1 : 0));
	}
}

// Finite Impulse Response Echo Filter
// 
// Filters the echo using an eight tap FIR filter:
// 
//         7
//        ---
//    x = \   c  * s
//        /    n    n
//        ---
//        n=0
// 
//    x = output sample
//    c = filter coefficient (-.7)
//    s = unfiltered sample
//    n = 0 is the oldest sample and 7 is the most recent
// 
// FIR filters are based on the sample rate.  This was fine in the SNES, because the sample rate was
// always 32kHz, but in the case of an emulator the sample rate can change.  So measures have to be
// taken to ensure the filter will have the same effect, regardless of the output sample rate.
// 
// To overcome this problem, I figured each tap of the filter is applied every 31250ns.  So the
// solution is to calculate when 31250ns have gone by, and use the sample at that point.  Of course
// this method really only works if the output rate is a multiple of 32k.  In order to get accurate
// results, some sort of interpolation method needs to be introduced.  I went the cheap route and used
// linear interpolation.
static inline void FIRFilter(int32 &l, int32 &r)
{
	int32 *fb = &Loop[(FIRBUF / 2) + firCur];

	// Move index back one sample. (Index will wrap around
	// after 64 samples, enough for up to 256kHz output.)
	firCur = (firCur + 2) & (FIRBUF / 2 - 1);

	// Store new samples in buffer
	fb[0] = fb[0 - (FIRBUF / 2)] = l;
	fb[1] = fb[1 - (FIRBUF / 2)] = r;

	l = 0;
	r = 0;

	// Reset decimal overflow, so filtering is consistant
	uint32 firDec = 0;

	for (int i = 7; i >= 0; i--) {
#if 0
		// linear interpolation
		l += (((((fb[-2] - fb[0]) * (int32)firDec) >> 16) + fb[0]) * FilterTaps[i]) >> 7;
		r += (((((fb[-1] - fb[1]) * (int32)firDec) >> 16) + fb[1]) * FilterTaps[i]) >> 7;
#else
		l += (fb[0] * FilterTaps[i]) >> 7;
		r += (fb[1] * FilterTaps[i]) >> 7;
#endif
		firDec += firRate;
		fb -= (firDec >> 16) * 2;
		firDec &= 0xffff;
	}

	//l = (int32)(int16)l & ~1;
	//r = (int32)(int16)r & ~1;
}

static inline int32 GetCurSample(int i)
{
	// 16bit sample data has the possibility of doubling
	// and therefore E_SHIFT is up to 8.

	switch(Settings.InterpolatedSound) {
	case INT_NONE:
		return (int32)mix[i].sBuf[mix[i].sIdx];

	case INT_LINEAR:
		{
			// Linear Interpolation
			int16 ps = mix[i].sBuf[(mix[i].sIdx - 1) & 0x1f];
			return (int32)((((((int32)mix[i].sBuf[mix[i].sIdx] - ps) >> 1) *
							 (int32)mix[i].mDec) >> 15) + ps);
		}

	case INT_CUBIC:
		{
			// Cubic Interpolation
			uint32 mDec = (mix[i].mDec >> 8) << 2;
			int32 cursample = (int32)(int16)((((int32)mix[i].sBuf[(mix[i].sIdx - 3) & 0x1f]
											   * cubicTab[mDec + 0]) >> 16)  //d0
											 +(((int32)mix[i].sBuf[(mix[i].sIdx - 2) & 0x1f]
												* cubicTab[mDec + 1]) >> 16)  //d1
											 +(((int32)mix[i].sBuf[(mix[i].sIdx - 1) & 0x1f]
												* cubicTab[mDec + 2]) >> 16)  //d2
											 +(((int32)mix[i].sBuf[mix[i].sIdx]
												* cubicTab[mDec + 3]) >> 16));//d3
			cursample += cursample;
			return cursample;
		}

	case INT_GAUSS:
		{
			// Gaussian Interpolation
			uint32 mDec = (mix[i].mDec >> 8) << 2;
			int32 cursample = (int32)(int16)((((int32)mix[i].sBuf[(mix[i].sIdx - 3) & 0x1f]
											   * gaussTab[mDec + 0]) >> 16)  //d0
											 +(((int32)mix[i].sBuf[(mix[i].sIdx - 2) & 0x1f]
												* gaussTab[mDec + 1]) >> 16)  //d1
											 +(((int32)mix[i].sBuf[(mix[i].sIdx - 1) & 0x1f]
												* gaussTab[mDec + 2]) >> 16)  //d2
											 +(((int32)mix[i].sBuf[mix[i].sIdx]
												* gaussTab[mDec + 3]) >> 16));//d3
			cursample += cursample;
			return cursample;
		}
	}
	return (int32)mix[i].sBuf[mix[i].sIdx];
}


//**************************************************************************************************
// Emulate DSP
//
// Emulates the DSP of the SNES
//
// Notes:
//    If 'pBuf' is NULL, the routine MIX_NONE will be used
//    Range checking is performed on 'size'
//
//    Callers should use EmuAPU instead
//
// In:
//    pBuf-> Buffer to store output
//	  num  = Length of buffer (in samples, can be 0)
//
// Out:
//    -> End of buffer
//
// Destroys:
//    ST(0-7)
void *EmuAPUDSP(void *pBuf, int32 num)
{
	uint8 *curptr = (uint8 *)pBuf;
	for(int32 mixlen = num; mixlen > 0; mixlen -= MIX_SIZE) {
		int32 len = mixlen > MIX_SIZE ? MIX_SIZE : mixlen;

		if(mixRout[(pBuf == NULL) ? MIX_NONE : dspMix](curptr, len))
			continue;

		// Output silence
		len = len * dspSize * dspChn;

		// mixlen is already in bytes instead of u32's, so no multiply is needed
		memset(curptr, (dspSize != 1) ? 0 : 0x80, len);
		curptr += len;
	}
	return curptr;
}

static bool8 EmuAPUDSPN(uint8 *&pBuf, int32 &num)
{
	int i, cnt;
	int si = 0;

	for(cnt = 0; cnt < num; cnt++, si += 2) {
		// Calculate Sample Counter
		t64Dec += t64Rate;
		t64CntS += (t64Dec >> 16);
		t64Dec &= 0xffff;

        // Generate Noise
		NoiseGen();

		// Erase current samples in buffer
		EchoBuffer[si + 0] = 0;
		EchoBuffer[si + 1] = 0;

		uint8 chm = 1;
		for(i = 0; i < 8; i++, chm <<= 1) {

			if(mix[i].mFlg & MFLG_CHG)
				MixingFlagProcess(i, chm);

			// Is the current voice active?
			if(!(mix[i].mFlg & MFLG_OFF)) {

				// Is there reproducing sound data?
				if(!(mix[i].mFlg & MFLG_END)) {
					// Pitch Modulation
					if(!(APU.DSP[APU_NON] & chm))
						PitchMod(i, chm);

					// Waveform Resizing
					mix[i].mDec += mix[i].mRate;
					uint32 c = (uint32)(mix[i].mDec >> 16);
					mix[i].mDec &= 0xffff;
					uint8 sIdx = (mix[i].sIdx + (c & 0x0f)) & 0x1f;
					// It needs next 16 samples?
					if((mix[i].sIdx ^ sIdx) & 0x10) {
						mix[i].sIdx = sIdx;
						UpdateSrc(i, chm);
					}
					else {
						mix[i].sIdx = sIdx;
					}
					if(!(mix[i].mFlg & MFLG_END)) {
						c >>= 4;
						while(c--) {
							mix[i].sIdx ^= 0x10;
							UpdateSrc(i, chm);
							if(mix[i].mFlg & MFLG_END)
								break;
						}
					}
				}

				int32 cursample;
				if(APU.DSP[APU_NON] & chm) {// Is noise enabled?
					cursample = mix[i].mOut = (nSmp * mix[i].eVal) >> (E_SHIFT + 7);
				}
				else if(!(mix[i].mFlg & MFLG_END)) {
					mix[i].mOut = ((int32)mix[i].sBuf[mix[i].sIdx] * mix[i].eVal) >> (E_SHIFT + 7);
					cursample = (GetCurSample(i) * mix[i].eVal) >> (E_SHIFT + 7);
				}
				else
					cursample = mix[i].mOut = 0;

				// Mixing
				if(cursample != 0 && !(mix[i].mFlg & MFLG_MUTE)) {// Not muted?
					if(APU.DSP[APU_EON] & chm){// Is echo on?
						EchoBuffer[si + 0] += (cursample * mix[i].mChnL) >> 7;
						EchoBuffer[si + 1] += (cursample * mix[i].mChnR) >> 7;
					}
				}

				// Envelope Calculation
				if(!(mix[i].eMode & E_IDLE)) {
					mix[i].eDec += mix[i].eRate;
					for(uint32 e = 0x10000; e <= mix[i].eDec; e += 0x10000) {
						UpdateEnv(i, chm);
						if(mix[i].eMode & E_IDLE)
							break;
					}
					mix[i].eDec &= 0xffff;
				}
			}
        }
    }

    if(!(APU.DSP[APU_FLG] & APU_MUTE)) {// Is the DSP not muted?
		si = 0;
		for(cnt = 0; cnt < num; cnt++, si += 2) {
			// Echo
			if(!disEcho) {
#ifdef NO_ESA
				int32 l = Echo[echoCur];
				int32 r = Echo[echoCur + 1];
#else
				int32 l = Echo[echoStart + echoCur];
				int32 r = Echo[echoStart + echoCur + 1];
#endif

				if(firEnabl) {
					// Filter echo
					FIRFilter(l, r);
				}

				// Add feedback
#ifdef NO_ESA
				Echo[echoCur] = ((l * echoFB) >> 7) + EchoBuffer[si + 0];
				Echo[echoCur + 1] = ((r * echoFB) >> 7) + EchoBuffer[si + 1];
#else
				Echo[echoStart + echoCur] = ((l * echoFB) >> 7) + EchoBuffer[si + 0];
				Echo[echoStart + echoCur + 1] = ((r * echoFB) >> 7) + EchoBuffer[si + 1];
#endif
				echoCur = (echoCur + 2) % echoDel;
			}
		}
	}
	return (pBuf == NULL);
}

// Emulate DSP (Integer)
//
// Emulates the DSP of the SNES using standard 80386 instructions.
// Samples can be 8 or 16-bit.
static bool8 EmuAPUDSPI(uint8 *&pBuf, int32 &num)
{
	int i, cnt;
	int si = 0;

	for(cnt = 0; cnt < num; cnt++, si += 2) {
		// Calculate Sample Counter
		t64Dec += t64Rate;
		t64CntS += (t64Dec >> 16);
		t64Dec &= 0xffff;

        // Generate Noise
		NoiseGen();

		// Erase current samples in buffer
		MixBuffer[si + 0] = 0;
		MixBuffer[si + 1] = 0;

		EchoBuffer[si + 0] = 0;
		EchoBuffer[si + 1] = 0;

		uint8 chm = 1;
		for(i = 0; i < 8; i++, chm <<= 1) {

			if(mix[i].mFlg & MFLG_CHG)
				MixingFlagProcess(i, chm);

			// Is the current voice active?
			if(!(mix[i].mFlg & MFLG_OFF)) {

				// Is there reproducing sound data?
				if(!(mix[i].mFlg & MFLG_END)) {
					// Pitch Modulation
					if(!(APU.DSP[APU_NON] & chm))
						PitchMod(i, chm);

					// Waveform Resizing
					mix[i].mDec += mix[i].mRate;
					uint32 c = (uint32)(mix[i].mDec >> 16);
					mix[i].mDec &= 0xffff;
					uint8 sIdx = (mix[i].sIdx + (c & 0x0f)) & 0x1f;
					// It needs next 16 samples?
					if((mix[i].sIdx ^ sIdx) & 0x10) {
						mix[i].sIdx = sIdx;
						UpdateSrc(i, chm);
					}
					else {
						mix[i].sIdx = sIdx;
					}
					if(!(mix[i].mFlg & MFLG_END)) {
						c >>= 4;
						while(c--) {
							mix[i].sIdx ^= 0x10;
							UpdateSrc(i, chm);
							if(mix[i].mFlg & MFLG_END)
								break;
						}
					}
				}

				int32 cursample;
				if(APU.DSP[APU_NON] & chm) {// Is noise enabled?
					cursample = mix[i].mOut = (nSmp * mix[i].eVal) >> (E_SHIFT + 7);
				}
				else if(!(mix[i].mFlg & MFLG_END)) {
					mix[i].mOut = ((int32)mix[i].sBuf[mix[i].sIdx] * mix[i].eVal) >> (E_SHIFT + 7);
					cursample = (GetCurSample(i) * mix[i].eVal) >> (E_SHIFT + 7);
				}
				else
					cursample = mix[i].mOut = 0;

				// Mixing
				if(cursample != 0 && !(mix[i].mFlg & MFLG_MUTE)) {// Not muted?
					MixBuffer[si + 0] += (cursample * mix[i].mChnL) >> 7;// Add to master samples
					MixBuffer[si + 1] += (cursample * mix[i].mChnR) >> 7;
					if(APU.DSP[APU_EON] & chm){// Is echo on?
						EchoBuffer[si + 0] += (cursample * mix[i].mChnL) >> 7;
						EchoBuffer[si + 1] += (cursample * mix[i].mChnR) >> 7;
					}
				}

				// Envelope Calculation
				if(!(mix[i].eMode & E_IDLE)) {
					mix[i].eDec += mix[i].eRate;
					for(uint32 e = 0x10000; e <= mix[i].eDec; e += 0x10000) {
						UpdateEnv(i, chm);
						if(mix[i].eMode & E_IDLE)
							break;
					}
					mix[i].eDec &= 0xffff;
				}
			}
        }
    }

    if(!(APU.DSP[APU_FLG] & APU_MUTE)) {// Is the DSP not muted?
		si = 0;
		int di = 0;
		uint8 *pBuf8 = (uint8 *)pBuf;
		int16 *pBuf16 = (int16 *)pBuf;
		for(cnt = 0; cnt < num; cnt++, si += 2, di += dspChn) {
			// Multiply samples by main volume
			MixBuffer[si + 0] = (MixBuffer[si + 0] * volMainL) >> 7;
			MixBuffer[si + 1] = (MixBuffer[si + 1] * volMainR) >> 7;

			int32 l, r;

			// Echo
			if(!disEcho) {
#ifdef NO_ESA
				l = Echo[echoCur];
				r = Echo[echoCur + 1];
#else
				l = Echo[echoStart + echoCur];
				r = Echo[echoStart + echoCur + 1];
#endif
				if(firEnabl) {
					// Filter echo
					FIRFilter(l, r);
				}

				// Add feedback
#ifdef NO_ESA
				Echo[echoCur] = ((l * echoFB) >> 7) + EchoBuffer[si + 0];
				Echo[echoCur + 1] = ((r * echoFB) >> 7) + EchoBuffer[si + 1];
#else
				Echo[echoStart + echoCur] = ((l * echoFB) >> 7) + EchoBuffer[si + 0];
				Echo[echoStart + echoCur + 1] = ((r * echoFB) >> 7) + EchoBuffer[si + 1];
#endif

				echoCur = (echoCur + 2) % echoDel;

				MixBuffer[si + 0] += (l * volEchoL) >> 7;
				MixBuffer[si + 1] += (r * volEchoR) >> 7;
			}
			
			// Render sound buffer
			switch(dspChn) {
			case 1:// mono
				switch(dspSize) {
				case 1:// 8bit
					l = (((MixBuffer[si + 0] + MixBuffer[si + 1]) * 181) >> 16) + 128;

					if(l < 0)
						l = 0;
					else if(l > 255)
						l = 255;

					pBuf8[di] = (uint8)l;
					break;

				case 2:// 16bit
					l = ((MixBuffer[si + 0] + MixBuffer[si + 1]) * 181) >> 8;

					if(l < -32768)
						l = -32768;
					else if(l > 32767)
						l = 32767;

					pBuf16[di] = (int16)l;
					break;

				default:
					return (pBuf == NULL);
				}
				break;

			case 2:// stereo
				switch(dspSize) {
				case 1:// 8bit
					l = (MixBuffer[si + 0] + 32768) >> 8;// Unsign sample
					r = (MixBuffer[si + 1] + 32768) >> 8;// Unsign sample

					if(l < 0)
						l = 0;
					else if(l > 255)
						l = 255;

					if(r < 0)
						r = 0;
					else if(r > 255)
						r = 255;

					pBuf8[di + 0] = (uint8)l;
					pBuf8[di + 1] = (uint8)r;
					break;

				case 2:// 16bit
					l = MixBuffer[si + 0];
					r = MixBuffer[si + 1];

					if(l < -32768)
						l = -32768;
					else if(l > 32767)
						l = 32767;


					if(r < -32768)
						r = -32768;
					else if(r > 32767)
						r = 32767;

					pBuf16[di + 0] = (int16)l;
					pBuf16[di + 1] = (int16)r;
					break;

				default:
					return (pBuf == NULL);
				}
				break;

			default:
				return (pBuf == NULL);
			}
		}

		if(pBuf) {
			switch(dspSize) {
			case 1:// 8bit
				pBuf = (uint8 *)(&pBuf8[di]);
				break;

			case 2:// 16bit
				pBuf = (uint8 *)(&pBuf16[di]);
				break;
			}
			return TRUE;
		}
	}
	return (pBuf == NULL);
}

static bool8 EmuAPUDSPX(uint8 *&pBuf, int32 &num)
{
	return EmuAPUDSPI(pBuf, num);
}

static bool8 EmuAPUDSPF(uint8 *&pBuf, int32 &num)
{
	return EmuAPUDSPI(pBuf, num);
}

//Decompress Sound Source
//
//Decompresses a 9-byte bit-rate reduced block into 16 16-bit samples
//
//In:
//  blk_hdr     = Block header
//  xsample_blk = Sample Block
//  output_buf  = Output buffer
//  smp_1       = Last sample of previous block
//  smp_2       = Next to last sample
//
//Out:
//  xsample_blk = Next Block
//  smp_1       = Last sample
//  smp_2       = Next to last sample
static void UnpckSrc(uint8 blk_hdr,
					 void *&xsample_blk,
					 int16 *output_buf,
					 int32 &smp_1,
					 int32 &smp_2)
{
	static uint8 buffer[9];
	uint8 *sample_blk = (uint8 *)xsample_blk;

	if(((uint8 *)sample_blk + 9) > end_pAPURAM) {
		int len = end_pAPURAM - sample_blk + 1;
		memcpy(buffer, sample_blk, len);
		memcpy(buffer + len, pAPURAM, 9 - len);
		xsample_blk = (uint8 *)sample_blk + 9 - 0x10000;
		sample_blk = buffer;
	}
	else
		xsample_blk = (uint8 *)sample_blk + 9;

	//skip header
	sample_blk++;
	int32 *BRR_row = &brrTab[(blk_hdr & 0xf0)];

    switch(blk_hdr & 0x0c){
	case 0x00:
        // Method0 - [Smp]
        for(int i = 0; i < 8; i++){
			output_buf[0] = BRR_row[sample_blk[0] >> 4];
			output_buf[1] = BRR_row[sample_blk[0] & 0x0f];
			sample_blk++;
			output_buf += 2;
        }
		smp_2 = output_buf[-2];
		smp_1 = output_buf[-1];
		break;

	case 0x04:
        // Method1 - [Delta]+[Smp-1](15/16)
        // [Smp-1] here is 16 bits, which means we have to truncate smp_1.
        for(int i = 0; i < 8; i++){
			smp_2 = (int32)(int16)(BRR_row[sample_blk[0] >> 4] +
								   smp_1 + (((-smp_1) >> 5) << 1));
			output_buf[0] = smp_2;

			smp_1 = (int32)(int16)(BRR_row[sample_blk[0] & 0x0f] +
								   smp_2 + (((-smp_2) >> 5) << 1));
			output_buf[1] = smp_1;

			sample_blk++;
			output_buf += 2;
        }
		break;

	case 0x08:
        // Method2 - [Delta]+[Smp-1](61/32)-[Smp-2](15/16)
        // [Smp-1] is something more than 16 bits, 24 seems good. [Smp-2] is 16
        // bits though. Pay attention!
        for(int i = 0; i < 8; i++){
			int32 smp;
			smp = ((smp_2 >> 5) << 1) - smp_2;
			smp_2 = smp_1;
			smp_1 = (BRR_row[sample_blk[0] >> 4] + (smp_1 << 1) +
					 ((-((smp_1 << 1) + smp_1) >> 6) << 1) + smp);
			if(smp_1 < -65536)
				smp_1 = 0;
			else if(smp_1 > 65534)
				smp_1 = -2;
			else
				smp_1 = (int32)(int16)smp_1;
			output_buf[0] = smp_1;

			smp = ((smp_2 >> 5) << 1) - smp_2;
			smp_2 = smp_1;
			smp_1 = (BRR_row[sample_blk[0] & 0x0f] + (smp_1 << 1) +
					 ((-((smp_1 << 1) + smp_1) >> 6) << 1) + smp);
			if(smp_1 < -65536)
				smp_1 = 0;
			else if(smp_1 > 65534)
				smp_1 = -2;
			else
				smp_1 = (int32)(int16)smp_1;
			output_buf[1] = smp_1;

			sample_blk++;
			output_buf += 2;
		}
		break;
 
	case 0x0c:
        // Method3 - [Delta]+[Smp-1](115/64)-[Smp-2](13/16)
        // [Smp-1] is something more than 16 bits, 24 seems good. [Smp-2] is 16
        // bits though. Pay attention!
        for(int i = 0; i < 8; i++) {
			int32 smp;
			smp = ((((smp_2 << 1) + smp_2) >> 5) << 1) - smp_2;
			smp_2 = smp_1;
			smp_1 = (BRR_row[sample_blk[0] >> 4] + (smp_1 << 1) +
					 ((-((smp_1 << 3) + (smp_1 << 2) + smp_1) >> 7) << 1) + smp);
			if(smp_1 < -65536)
				smp_1 = 0;
			else if(smp_1 > 65534)
				smp_1 = -2;
			else
				smp_1 = (int32)(int16)smp_1;
			output_buf[0] = smp_1;

			smp = ((((smp_2 << 1) + smp_2) >> 5) << 1) - smp_2;
			smp_2 = smp_1;
			smp_1 = (BRR_row[sample_blk[0] & 0x0f] + (smp_1 << 1) +
					 ((-((smp_1 << 3) + (smp_1 << 2) + smp_1) >> 7) << 1) + smp);
			if(smp_1 < -65536)
				smp_1 = 0;
			else if(smp_1 > 65534)
				smp_1 = -2;
			else
				smp_1 = (int32)(int16)smp_1;
			output_buf[1] = smp_1;

			sample_blk++;
			output_buf += 2;
        }
		break;
    }
}

